# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
r"""
Set of functions for internal module use.
"""

import torch

# SVD decomposition
# Default behavior: vanilla svd
_IS_TORCH_BATCH_SVD_AVAILABLE = False
try:
    # Should raise an AttributeError exception if undefined.
    torch.linalg.svd
    def svd(M):
        r"""
        Singular Value Decomposition wrapper.
        
        Args:
            M (BxMxN tensor): batch of real matrices.
        Returns:
            (U,D,V) decomposition, such as :math:`M = U @ diag(D) @ V^T`.
        """
        U, D, Vt = torch.linalg.svd(M)
        return (U, D, Vt.transpose(-2,-1))
except (NameError, AttributeError):
    # deprecated in torch 2.0
    svd = torch.svd

# With PyTorch < 1.8,
# we observed some significant speed-ups using torch_batch_svd (https://github.com/KinglittleQ/torch-batch-svd) instead of torch.svd on NVidia GPUs.
# In more recent versions, this is no longer required (https://github.com/pytorch/pytorch/pull/48436).
_torch_version_major, _torch_version_minor = [int(s) for s in torch.__version__.split(".")[:2]]
if  _torch_version_major == 0 or (_torch_version_major == 1 and _torch_version_minor < 8):
    try:
        import torch_batch_svd
        _IS_TORCH_BATCH_SVD_AVAILABLE = True
        def svd(M):
            r"""
            Singular Value Decomposition wrapper, using efficient batch decomposition on GPU.

            Args:
                M (BxMxN tensor): batch of real matrices.
            Returns:
                (U,D,V) decomposition, such as :math:`M = U @ diag(D) @ V^T`.
            """
            if M.is_cuda and M.shape[1] < 32 and M.shape[2] < 32:
                return torch_batch_svd.svd(M)
            else:
                return torch.svd(M)
    except ModuleNotFoundError:
        pass
del _torch_version_major, _torch_version_minor

def flatten_batch_dims(tensor, end_dim):
    r"""
    :meta private:
    Utility function: flatten multiple batch dimensions into a single one, or add a batch dimension if there is none.
    """
    batch_shape = tensor.shape[:end_dim+1]
    flattened = tensor.flatten(end_dim=end_dim) if len(batch_shape) > 0 else tensor.unsqueeze(0)
    return flattened, batch_shape

def unflatten_batch_dims(tensor, batch_shape):
    r"""
    :meta private:
    Revert flattening of a tensor.
    """
    # Note: alternative to tensor.unflatten(dim=0, sizes=batch_shape) that was not supported by PyTorch 1.6.0.
    return tensor.reshape(batch_shape + tensor.shape[1:]) if len(batch_shape) > 0 else tensor.squeeze(0)

def _pseudo_inverse(x, eps):
    r"""
    :meta private:
    Element-wise pseudo inverse.
    """
    inv = 1.0/x
    inv[torch.abs(x) < eps] = 0.0
    return inv    

# Batched eigenvalue decomposition.
# Recent version of PyTorch deprecated the use of torch.symeig.
try:
    torch.linalg.eigh
    def symeig_lower(A):
        r"""
        Batched eigenvalue decomposition. Only the lower part of the matrix is considered.
        """
        return torch.linalg.eigh(A, UPLO='L')
except (NameError, AttributeError):
    # Older PyTorch version
    def symeig_lower(A):
        r"""
        Batched eigenvalue decomposition. Only the lower part of the matrix is considered.
        """
        return torch.symeig(A, upper=False, eigenvectors=True)
    
# L2 normalization
try:
    torch.linalg.norm
    def norm(x, dim=None, keepdim=False):
        return torch.linalg.norm(x, dim=dim, keepdim=keepdim)
except AttributeError:
    # torch.linalg.norm was introduced in PyTorch 1.7, and torch.norm is deprecated.
    def norm(x, dim=None, keepdim=False):
        return torch.norm(x, dim=dim, keepdim=keepdim)
    
try:
    torch.hypot
    hypot = torch.hypot
except AttributeError:
    # torch.hypot is not available in PyTorch 1.6.
    def hypot(x, y):
        return torch.sqrt(torch.square(x) + torch.square(y))

# torch.amp.custom_fwd and torch.amp.custom_bwd were not available in PyTorch 1.6.
try:
    custom_fwd = torch.amp.custom_fwd
except AttributeError:
    def custom_fwd(**kwargs):
        return lambda x : x
try:
    custom_bwd = torch.amp.custom_bwd
except AttributeError:
    def custom_bwd(**kwargs):
        return lambda x : x



# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import torch
import roma
import numpy as np
import roma.internal

def _elementary_basis_index(axis):
    r"""
    Return the index corresponding to a given axis label.
    """
    if axis == 'x':
        return 0
    elif axis == 'y':
        return 1
    elif axis == 'z':
        return 2
    else:
        raise ValueError("Invalid axis.")

def euler_to_unitquat(convention: str, angles, degrees=False, normalize=True, dtype=None, device=None):
    r"""
    Convert Euler angles to unit quaternion representation.

    Args:
        convention (string): string defining a sequence of D rotation axes ('XYZ' or 'xzx' for example).
            The sequence of rotation is expressed either with respect to a global 'extrinsic' coordinate system (in which case axes are denoted in lowercase: 'x', 'y', or 'z'),
            or with respect to an 'intrinsic' coordinates system attached to the object under rotation (in which case axes are denoted in uppercase: 'X', 'Y', 'Z').
            Intrinsic and extrinsic conventions cannot be mixed.
        angles (...xD tensor, or tuple/list of D floats or ... tensors): a list of angles associated to each axis, expressed in radians by default.
        degrees (bool): if True, input angles are assumed to be expressed in degrees.
        normalize (bool): if True, normalize the returned quaternion to compensate potential numerical.
    
    Returns:
        A batch of unit quaternions (...x4 tensor, XYZW convention).

    Warning:
        Case is important: 'xyz' and 'XYZ' denote different conventions.
    """
    if type(angles) == torch.Tensor:
        angles = [t.squeeze(dim=-1) for t in torch.split(angles, split_size_or_sections=1, dim=-1)]

    assert len(convention) == len(angles)

    extrinsics = convention.islower()
    if extrinsics:
        # Cast from intrinsics to extrinsics convention
        convention = convention.upper()[::-1]
        angles = angles[::-1]

    unitquats = []
    for axis, angle in zip(convention, angles):
        angle = torch.as_tensor(angle, device=device, dtype=dtype)
        if degrees:
            angle = torch.deg2rad(angle)
        batch_shape = angle.shape
        rotvec = torch.zeros(batch_shape + torch.Size((3,)), device=angle.device, dtype=angle.dtype)
        if axis == 'X':
            rotvec[...,0] = angle
        elif axis == 'Y':
            rotvec[...,1] = angle
        elif axis == 'Z':
            rotvec[...,2] = angle
        else:
            raise ValueError("Invalid convention (expected format: 'xyz', 'zxz', 'XYZ', etc.).")
        q = roma.rotvec_to_unitquat(rotvec)
        unitquats.append(q)
    if len(unitquats) == 1:
        return unitquats[0]
    else:
        return roma.quat_composition(unitquats, normalize=normalize)

def euler_to_rotvec(convention: str, angles, degrees=False, dtype=None, device=None):
    r"""
    Convert Euler angles to rotation vector representation.

    Args:
        convention (string): 'xyz' for example. See :func:`~roma.euler.euler_to_unitquat()`.
        angles (...xD tensor, or tuple/list of D floats or ... tensors): a list of angles associated to each axis, expressed in radians by default.
        degrees (bool): if True, input angles are assumed to be expressed in degrees.

    Returns:
        a batch of rotation vectors (...x3 tensor).
    """
    return roma.unitquat_to_rotvec(euler_to_unitquat(convention=convention, angles=angles, degrees=degrees, dtype=dtype, device=device))

def euler_to_rotmat(convention: str, angles, degrees=False, dtype=None, device=None):
    r"""
    Convert Euler angles to rotation matrix representation.

    Args:
        convention (string): 'xyz' for example. See :func:`~roma.euler.euler_to_unitquat()`.
        angles (...xD tensor, or tuple/list of D floats or ... tensors): a list of angles associated to each axis, expressed in radians by default.
        degrees (bool): if True, input angles are assumed to be expressed in degrees.
    
    Returns:
        a batch of rotation matrices (...x3x3 tensor).
    """
    return roma.unitquat_to_rotmat(euler_to_unitquat(convention=convention, angles=angles, degrees=degrees, dtype=dtype, device=device))

def unitquat_to_euler(convention : str, quat, as_tuple=False, degrees=False, epsilon=1e-7):
    r"""
    Convert unit quaternion to Euler angles representation.

    Args:
        convention (str): string of 3 characters belonging to {'x', 'y', 'z'} for extrinsic rotations, or {'X', 'Y', 'Z'} for intrinsic rotations.
            Consecutive axes should not be identical.
        quat (...x4 tensor, XYZW convention): input batch of unit quaternion.
        as_tuple (boolean): if True, angles are not stacked but returned as a tuple of tensors.
        degrees (bool): if True, angles are returned in degrees.
        epsilon (float): a small value used to detect degenerate configurations.

    Returns:
        A stacked ...x3 tensor corresponding to Euler angles, expressed by default in radians.
        In case of gimbal lock, the third angle is arbitrarily set to 0.
    """
    # Code adapted from scipy.spatial.transform.Rotation.
    # Reference: https://github.com/scipy/scipy/blob/ac6bcaf00411286271f7cc21e495192c73168ae4/scipy/spatial/transform/_rotation.pyx#L325C12-L325C15
    assert len(convention) == 3

    pi = np.pi
    lamb = np.pi/2

    extrinsic = convention.islower()
    if not extrinsic:
        convention = convention.lower()[::-1]

    quat, batch_shape = roma.internal.flatten_batch_dims(quat, end_dim=-2)
    N = quat.shape[0]

    i = _elementary_basis_index(convention[0])
    j = _elementary_basis_index(convention[1])
    k = _elementary_basis_index(convention[2])
    assert i != j and j != k, "Consecutive axes should not be identical."

    symmetric = (i == k)

    if symmetric:
        # Get third axis
        k = 3 - i - j

    # Step 0
    # Check if permutation is even (+1) or odd (-1) 
    sign = (i - j) * (j - k) * (k - i) // 2

    # Step 1
    # Permutate quaternion elements            
    if symmetric:
        a = quat[:,3]
        b = quat[:,i]
        c = quat[:,j]
        d = quat[:,k] * sign
    else:
        a = quat[:,3] - quat[:,j]
        b = quat[:,i] + quat[:,k] * sign
        c = quat[:,j] + quat[:,3]
        d = quat[:,k] * sign - quat[:,i]

    
    # intrinsic/extrinsic conversion helpers
    if extrinsic:
        angle_first = 0
        angle_third = 2
    else:
        angle_first = 2
        angle_third = 0

    # Step 2
    # Compute second angle...
    angles = [torch.empty(N, device=quat.device, dtype=quat.dtype) for _ in range(3)]
    
    angles[1] = 2 * torch.atan2(roma.internal.hypot(c, d), roma.internal.hypot(a, b))

    # ... and check if equal to is 0 or pi, causing a singularity
    case1 = torch.abs(angles[1]) <= epsilon
    case2 = torch.abs(angles[1] - pi) <= epsilon
    case1or2 = torch.logical_or(case1, case2)
    # Step 3
    # compute first and third angles, according to case
    half_sum = torch.atan2(b, a)
    half_diff = torch.atan2(d, c)
    
    # no singularities
    angles[angle_first] = half_sum - half_diff
    angles[angle_third] = half_sum + half_diff
    
    # any degenerate case
    angles[2][case1or2] = 0
    angles[0][case1] = 2 * half_sum[case1]
    angles[0][case2] = 2 * (-1 if extrinsic else 1) * half_diff[case2]
            
    # for Tait-Bryan/asymmetric sequences
    if not symmetric:
        angles[angle_third] *= sign
        angles[1] -= lamb

    for idx in range(3):
        foo = angles[idx]
        foo[foo < -pi] += 2 * pi
        foo[foo > pi] -= 2 * pi
        if degrees:
            foo = torch.rad2deg(foo)
        angles[idx] = roma.internal.unflatten_batch_dims(foo, batch_shape)

    if as_tuple:
        return tuple(angles)
    else:
        return torch.stack(angles, dim=-1)

def rotvec_to_euler(convention : str, rotvec, as_tuple=False, degrees=False, epsilon=1e-7):
    r"""
    Convert rotation vector to Euler angles representation.

    Args:
        convention (str): string of 3 characters belonging to {'x', 'y', 'z'} for extrinsic rotations, or {'X', 'Y', 'Z'} for intrinsic rotations.
            Consecutive axes should not be identical.
        rotvec (...x3 tensor): input batch of rotation vectors.
        as_tuple (boolean): if True, angles are not stacked but returned as a tuple of tensors.
        degrees (bool): if True, angles are returned in degrees.
        epsilon (float): a small value used to detect degenerate configurations.

    Returns:
        A stacked ...x3 tensor corresponding to Euler angles, expressed by default in radians.
        In case of gimbal lock, the third angle is arbitrarily set to 0.
    """
    return unitquat_to_euler(convention, roma.rotvec_to_unitquat(rotvec), as_tuple=as_tuple, degrees=degrees, epsilon=epsilon)

def rotmat_to_euler(convention : str, rotmat, as_tuple=False, degrees=False, epsilon=1e-7):
    r"""
    Convert rotation matrix to Euler angles representation.

    Args:
        convention (str): string of 3 characters belonging to {'x', 'y', 'z'} for extrinsic rotations, or {'X', 'Y', 'Z'} for intrinsic rotations.
            Consecutive axes should not be identical.
        rotmat (...x3x3 tensor): input batch of rotation matrices.
        as_tuple (boolean): if True, angles are not stacked but returned as a tuple of tensors.
        degrees (bool): if True, angles are returned in degrees.
        epsilon (float): a small value used to detect degenerate configurations.

    Returns:
        A stacked ...x3 tensor corresponding to Euler angles, expressed by default in radians.
        In case of gimbal lock, the third angle is arbitrarily set to 0.
    """
    return unitquat_to_euler(convention, roma.rotmat_to_unitquat(rotmat), as_tuple=as_tuple, degrees=degrees, epsilon=epsilon)


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import unittest
import torch
import numpy as np
import roma
import roma.internal
from test.utils import is_close

class TestUtils(unittest.TestCase):
    def test_flatten(self):
        for dtype in (torch.float32, torch.float64):
            x = torch.randn((32, 24, 3, 4), dtype=dtype)
            xflat, batch_shape = roma.internal.flatten_batch_dims(x, end_dim=-3)
            self.assertEqual(batch_shape, (32, 24))
            self.assertTrue(xflat.shape == (32*24, 3, 4))
            xbis = roma.internal.unflatten_batch_dims(xflat, batch_shape)
            self.assertTrue(torch.all(xbis == x))

    def test_rotmat_geodesic_distance(self):
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            axis = torch.nn.functional.normalize(torch.randn((batch_size,3), dtype=dtype), dim=-1)
            alpha = (np.pi - 1e-5) * 2 * (torch.rand(batch_size, dtype=dtype)-0.5)
            x = alpha[:,None] * axis
            R = roma.rotvec_to_rotmat(x)
            
            cosine = roma.rotmat_cosine_angle(R)
            self.assertTrue(is_close(cosine, torch.cos(alpha)))
            
            I = torch.eye(3, dtype=dtype)
            M = roma.random_rotmat(batch_size, dtype=dtype)
            
            geo_dist = roma.rotmat_geodesic_distance(R, I[None,:,:])
            self.assertTrue(is_close(torch.abs(alpha), geo_dist))
            
            # Left-invariance of the metric
            geo_dist_bis = roma.rotmat_geodesic_distance(M @ R, M @ I[None,:,:])
            self.assertTrue(is_close(geo_dist_bis, geo_dist))
            
            # Right-invariance of the metric
            geo_dist_ter = roma.rotmat_geodesic_distance(R @ M, I[None,:,:] @ M)
            self.assertTrue(is_close(geo_dist_ter, geo_dist))
            
            geo_dist_naive = roma.rotmat_geodesic_distance_naive(M @ R, M @ I[None,:,:])
            self.assertTrue(is_close(torch.abs(alpha), geo_dist_naive))

    def test_other_geodesic_distance(self):
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            q1 = roma.random_unitquat(batch_size, dtype=dtype)
            q2 = roma.random_unitquat(batch_size, dtype=dtype)
            alpha_q = roma.unitquat_geodesic_distance(q1, q2)
            # Ensure consistency between functions
            R1 = roma.unitquat_to_rotmat(q1)
            R2 = roma.unitquat_to_rotmat(q2)
            alpha_R = roma.rotmat_geodesic_distance(R1, R2)
            self.assertTrue(is_close(alpha_q, alpha_R))
            rotvec1 = roma.unitquat_to_rotvec(q1)
            rotvec2 = roma.unitquat_to_rotvec(q2)
            alpha_rotvec = roma.rotvec_geodesic_distance(rotvec1, rotvec2)
            self.assertTrue(is_close(alpha_rotvec, alpha_q))

    def test_identity_quat(self):
        q = roma.identity_quat()
        self.assertTrue(q.shape == (4,))
        self.assertTrue(is_close(q, roma.quat_inverse(q)))

        q = roma.identity_quat(5)
        self.assertTrue(q.shape == (5,4))
        self.assertTrue(is_close(q, roma.quat_inverse(q)))

        q = roma.identity_quat((3,2))
        self.assertTrue(q.shape == (3,2,4))
        self.assertTrue(is_close(q, roma.quat_inverse(q)))

    def test_random_unitquat(self):
        q = roma.random_unitquat((3,5))
        self.assertTrue(q.shape == (3,5,4))

        batch_size = 100000
        repetitions = 10
        torch.manual_seed(565441)
        for _ in range(10):
            for dtype in (torch.float32, torch.float64):
                # Sample multiple times a batch of random quaternions
                q = torch.cat([roma.random_unitquat(batch_size, dtype=dtype) for _ in range(repetitions)], dim=0)
                size = repetitions * batch_size
                self.assertTrue(q.shape == (size, 4))

                # Ensure that the generated quaternions are of unit norm
                self.assertTrue(torch.all(torch.abs(torch.norm(q, dim=-1) - 1.0) < 1e-6))

                # If the distribution is uniform, the mean quaternion should be close to 0
                # (we use a constant random seed for this test, otherwise it could fail in some unlikely cases).
                mean = torch.mean(q, dim=0)
                self.assertTrue(torch.norm(mean) < 10.0 / np.sqrt(size))

                # There should be approximately as many quaternions on both sides of a random hyperplane going through 0.
                for _ in range(5):
                    axis = torch.randn((size, 4), dtype=dtype)
                    axis /= torch.norm(axis, dim=-1, keepdim=True)
                    count_difference = torch.sum(torch.sign(torch.sum(axis * q, dim=-1)))
                    self.assertTrue(torch.abs(count_difference) < 10.0 * np.sqrt(size))

    def test_unitquat(self):
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            q = roma.random_unitquat(batch_size, dtype=dtype)
            # Ensure that conjugation of a unit quaternion equates to its inverse.
            iq = roma.quat_conjugation(q)
            q_id = roma.rotvec_to_unitquat(torch.zeros(1,3))
            self.assertTrue(is_close(q_id, roma.quat_product(q, iq)))
            self.assertTrue(is_close(q_id, roma.quat_product(iq, q)))

    def test_quat(self):
        batch_size = 100
        for dtype in (torch.float32, torch.float64):
            q = roma.random_unitquat(batch_size, dtype=dtype)
            # Test quaternion inverse (hoping for non zero quaternions during the test).
            iq = roma.quat_inverse(q)
            q_id = roma.rotvec_to_unitquat(torch.zeros(1,3))
            self.assertTrue(is_close(q_id, roma.quat_product(q, iq)))
            self.assertTrue(is_close(q_id, roma.quat_product(iq, q)))
            nq = roma.quat_normalize(q)
            self.assertTrue(is_close(roma.internal.norm(nq, dim=-1), torch.ones(batch_size, dtype=dtype)))

    def test_quat_action(self):
        batch_size=100
        for dtype in (torch.float32, torch.float64):
            q = torch.randn(batch_size, 4)
            v = torch.randn(batch_size, 3)
            vrot0 = roma.quat_action(q, v)

            # Check that results are consistent with respect to quaternion normalization
            q_normalized = q / torch.norm(q, dim=-1, keepdim=True)
            vrot1 = roma.quat_action(q_normalized, v, is_normalized=True)
            vrot2 = roma.quat_action(q_normalized, v, is_normalized=False)

            self.assertTrue(is_close(vrot0, vrot1))
            self.assertTrue(is_close(vrot1, vrot2))

            # Check that results are consistent with rotation using a rotation matrix
            R = roma.unitquat_to_rotmat(q_normalized)
            vrotR = torch.einsum('bik, bk -> bi', R, v)
            self.assertTrue(is_close(vrot1, vrotR))
        
    def test_slerp(self):
        for dtype in (torch.float32, torch.float64):
            for batch_shape in [(100,), (24, 32), (3, 4, 7)]:
                # Identity
                x0 = torch.zeros(1, 3, dtype=dtype)
                # 'Small' rotation vector
                axis = torch.nn.functional.normalize(torch.randn(batch_shape + (3,), dtype=dtype), dim=-1)
                alpha = (np.pi-1e-10) * 2 * (torch.rand(batch_shape, dtype=dtype)-0.5)
                x1 = alpha[...,None] * axis
                
                # Rotvec slerp
                steps = torch.linspace(0, 1, 11, dtype=dtype)
                x_true = steps.reshape(-1, *([1] * (len(batch_shape) + 1))) * x1.unsqueeze(0)
                x_pred = roma.rotvec_slerp(x0.expand(batch_shape + (3,)), x1, steps)
                self.assertTrue(is_close(x_true, x_pred))
                
                # Check left invariance of unit quat slerp
                q0 = roma.random_unitquat(batch_shape, dtype=dtype)
                q1 = roma.random_unitquat(batch_shape, dtype=dtype)
                q = roma.random_unitquat(batch_shape, dtype=dtype)
                iq = roma.quat_conjugation(q)
                slerp1 = roma.unitquat_slerp(q0, q1, steps)
                slerp2 = roma.quat_product(iq.expand(len(steps), *batch_shape, 4),
                                        roma.unitquat_slerp(roma.quat_product(q, q0),
                                                                roma.quat_product(q, q1),
                                                                steps))
                self.assertTrue(is_close(slerp1, slerp2))

                # Check consistency between rotmat_slerp and unitquat_slerp
                R0 = roma.unitquat_to_rotmat(q0)
                R1 = roma.unitquat_to_rotmat(q1)
                R_slerp = roma.rotmat_slerp(R0, R1, steps)
                self.assertTrue(is_close(roma.unitquat_to_rotmat(slerp1), R_slerp))

                # Check that the first and last element of quaternion slerp are consistent with the inputs
                q = roma.unitquat_slerp(q0, q1, steps, shortest_arc=False)
                self.assertTrue(is_close(q[0], q0))
                self.assertTrue(is_close(q[-1], q1))

                # Similar test when shortest_arc==True
                q = roma.unitquat_slerp(q0, q1, steps, shortest_arc=True)
                self.assertTrue(is_close(q[0], q0))
                self.assertTrue(torch.all(torch.min(torch.norm(q[-1] - q1, dim=-1), torch.norm(q[-1] + q1, dim=-1)) < 5e-6))

    def test_unitquat_slerp(self):
        # Test of slerp on a specific example,
        # where enforcing -- or not -- interpolation along the shortest arc give different results.
        q0 = torch.as_tensor([1,0,0,0], dtype=torch.float)
        q1 = -torch.as_tensor([1,1,0,0], dtype=torch.float)
        q1 /= torch.norm(q1, keepdim=True)
        steps = torch.linspace(0, 1.0, 3)

        q05 = q0 + q1
        q05 /= torch.norm(q05, keepdim=True)
        q = roma.utils.unitquat_slerp(q0, q1, steps, shortest_arc=False)
        self.assertTrue(is_close(q, torch.stack((q0, q05, q1))))

        # Interpolation along the shortest arc
        q05m = q0 - q1
        q05m /= torch.norm(q05m, keepdim=True)
        q = roma.utils.unitquat_slerp(q0, q1, steps, shortest_arc=True)
        self.assertTrue(is_close(q, torch.stack((q0, q05m, -q1))))

    def test_slerp_consistency(self):
        # Test consistency between both slerp methods.
        for batch_shape in [(3,), (10, 20), (3, 14, 7)]:
            for steps_shape in [(6,), (4,2,), (3,2,1)]:
                for shortest_arc in (True, False):
                    q0 = roma.random_unitquat(batch_shape)
                    q1 = roma.random_unitquat(batch_shape)
                    steps = torch.rand(steps_shape)
                    q = roma.unitquat_slerp(q0, q1, steps=steps, shortest_arc=shortest_arc)
                    qbis = roma.unitquat_slerp_fast(q0, q1, steps=steps, shortest_arc=shortest_arc)
                    self.assertTrue(q.shape == steps_shape + batch_shape + (4,))
                    self.assertTrue(is_close(q, qbis))
                    # Same tests with nearby rotations
                    q1 = q0 + 1e-3 * torch.randn_like(q0)
                    q1 /= torch.norm(q1, dim=-1, keepdim=True)
                    q = roma.unitquat_slerp(q0, q1, steps=steps, shortest_arc=shortest_arc)
                    qbis = roma.unitquat_slerp_fast(q0, q1, steps=steps, shortest_arc=shortest_arc)
                    self.assertTrue(q.shape == steps_shape + batch_shape + (4,))
                    self.assertTrue(is_close(q, qbis))            

    def test_composition(self):
        for dtype in (torch.float32, torch.float64):
            R_list = [roma.random_rotmat(dtype=dtype) for _ in range(3)]
            q_list = [roma.rotmat_to_unitquat(R) for R in R_list]
            rotvec_list = [roma.rotmat_to_rotvec(R) for R in R_list]
            
            # Test consistency of results between the different representations
            R = roma.rotmat_composition(R_list)
            q = roma.quat_composition(q_list)
            rotvec = roma.rotvec_composition(rotvec_list)
            self.assertTrue(is_close(roma.unitquat_to_rotmat(q), R))
            self.assertTrue(is_close(roma.rotvec_to_rotmat(rotvec), R))

            # Test consistency with respect to inverse
            self.assertTrue(is_close(roma.rotmat_cosine_angle(roma.rotmat_composition((R_list[0], roma.rotmat_inverse(R_list[0])))), torch.as_tensor(1.0)))
            self.assertTrue(is_close(roma.quat_composition((q_list[0], roma.quat_inverse(q_list[0]))), torch.as_tensor([0.0, 0.0, 0.0, 1])))
            self.assertTrue(is_close(roma.rotvec_composition((rotvec_list[0], roma.rotvec_inverse(rotvec_list[0]))), torch.zeros(3)))

    def test_rigid_vectors_registration(self):
        batch_shape = (34, 16)
        n = 100
        for dtype in (torch.float32, torch.float64):
            for weights in [None, torch.rand(size=batch_shape + (n,), dtype=dtype)]:
                R_true = roma.random_rotmat(batch_shape, dtype=dtype)
                X = torch.randn(batch_shape + (n, 3,), dtype=dtype)
                Y = torch.einsum('...ik, ...jk -> ...ji', R_true, X)
                R = roma.rigid_vectors_registration(X, Y, weights)
                self.assertTrue(is_close(R, R_true))

    def test_rigid_vectors_registration_with_scale(self):
        batch_shape = (34, 16)
        n = 100
        for dtype in (torch.float32, torch.float64):
            for weights in [None, torch.rand(size=batch_shape + (n,), dtype=dtype)]:
                R_true = roma.random_rotmat(batch_shape, dtype=dtype)
                scale_true = torch.exp(torch.randn(size=batch_shape, dtype=dtype))
                X = torch.randn(batch_shape + (n, 3,), dtype=dtype)
                Y = scale_true[...,None, None] * torch.einsum('...ik, ...jk -> ...ji', R_true, X)
                R, scale = roma.rigid_vectors_registration(X, Y, weights, compute_scaling=True)
                self.assertTrue(is_close(R, R_true))
                self.assertTrue(is_close(scale, scale_true))

    def test_rigid_point_registration(self):
        batch_shape = (34, 16)
        n = 100
        for dtype in (torch.float32, torch.float64):
            for weights in [None, torch.rand(size=batch_shape + (n,), dtype=dtype)]:
                R_true = roma.random_rotmat(batch_shape, dtype=dtype)
                t_true = torch.randn(batch_shape + (3,), dtype=dtype)
                X = torch.randn(batch_shape + (n, 3,), dtype=dtype)
                Y = torch.einsum('...ik, ...jk -> ...ji', R_true, X) + t_true.unsqueeze(-2)
                R, t = roma.rigid_points_registration(X, Y, weights)

                self.assertTrue(is_close(R, R_true))
                self.assertTrue(is_close(t, t_true))

    def test_rigid_points_registration_with_scale(self):
        batch_shape = (34, 16)
        n = 100
        for dtype in (torch.float32, torch.float64):
            for weights in [None, torch.rand(size=batch_shape + (n,), dtype=dtype)]:
                R_true = roma.random_rotmat(batch_shape, dtype=dtype)
                t_true = torch.randn(batch_shape + (3,), dtype=dtype)
                scale_true = torch.exp(torch.randn(size=batch_shape, dtype=dtype))
                X = torch.randn(batch_shape + (n, 3,), dtype=dtype)
                Y = scale_true[...,None, None] * torch.einsum('...ik, ...jk -> ...ji', R_true, X) + t_true.unsqueeze(-2)
                R, t, scale = roma.rigid_points_registration(X, Y, weights, compute_scaling=True)
                self.assertTrue(is_close(R, R_true))
                self.assertTrue(is_close(t, t_true))
                self.assertTrue(is_close(scale, scale_true))
        
if __name__ == "__main__":
    unittest.main()
    


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
r"""
Spatial transformations parameterized by rotation matrices, unit quaternions and more.

Example of use
~~~~~~~~~~~~~~

.. literalinclude :: ../../examples/snippets/transforms.py
    :language: python

.. _apply-transformation:

Applying a transformation
~~~~~~~~~~~~~~~~~~~~~~~~~

When applying a transformation to a set of points of coordinates :code:`v`,
The batch shape of :code:`v` should be broadcastable with the batch shape of the transformation.

For example, one can sample a unique random rigid 3D transformation and use it to transform 100 random 3D points as follows:

.. code-block:: python

    roma.Rigid(roma.random_rotmat(), torch.randn(3))[None].apply(torch.randn(100,3))

To apply a different transformation to each point, one could use instead:

.. code-block:: python

    roma.Rigid(roma.random_rotmat(100), torch.randn(100,3)).apply(torch.randn(100,3))

.. _aliasing_issues:    

Aliasing issues
~~~~~~~~~~~~~~~

.. warning::

    For efficiency reasons, transformation objects do not copy input data. Be careful if you intend to do some in-place data modifications, and use the :code:`clone()` method when required.

"""
import torch
import roma
import warnings
 
class Linear:
    r"""
    A linear transformation parameterized by a matrix :math:`M \in \mathcal{M}_{D,C}(\mathbb{R})`,
    transforming a point :math:`x \in \mathbb{R}^C` into :math:`M x`.

    :var linear: (...xDxC tensor): batch of matrices specifying the transformations considered.
    """
    def __init__(self, linear):
        self.linear = linear

    def linear_compose(self, other):
        r"""
        Compose the linear part of two transformations.

        Args:
            other: an other transformation of same type.
        Returns:
            a tensor representing the composed transformation.
        """
        assert len(self.linear.shape) == len(other.linear.shape), "Expecting the same number of batch dimensions for the two transformations."
        return torch.einsum("...ik, ...kj -> ...ij", self.linear, other.linear)
    
    def linear_inverse(self):
        r"""
        Returns:
            The inverse of the linear transformation, when applicable.
        """
        return torch.inverse(self.linear)
    
    def linear_apply(self, v):
        r"""
        Transforms a tensor of vector coordinates.

        Args:
            v (...xD tensor): tensor of vector coordinates to transform.

        Returns:
            The transformed vector coordinates.

        See note in :func:`~roma.transforms.Linear.apply()` regarding broadcasting.
        """
        assert len(self.linear.shape) == len(v.shape) + 1, "Expecting the same number of batch dimensions for the transformation and the vector."
        return torch.einsum("...ik, ...k -> ...i", self.linear, v)

    def linear_normalize(self):
        return self.linear
    
    def compose(self, other):
        r"""
        Compose a transformation with the current one.

        Args:
            other: an other transformation of same type.

        Returns:
            The resulting transformation.
        """
        return type(self)(self.linear_compose(other))
    
    def inverse(self):
        r"""
        Returns:
            The inverse transformation, when applicable.
        """
        return type(self)(self.linear_inverse())
    
    def apply(self, v):
        r"""
        Transforms a tensor of points coordinates. See :ref:`apply-transformation`.

        Args:
            v (...xD tensor): tensor of point coordinates to transform.

        Returns:
            The transformed point coordinates.
        """
        return self.linear_apply(v)
    
    def normalize(self):
        r"""
        Returns:
            Copy of the transformation, normalized to ensure the class properties
            (for example to ensure that a :class:`Rotation` object is an actual rotation).
        """
        return type(self)(self.linear_normalize())
    
    def __matmul__(self, other):
        r"""
        Overloading of the `@` operator for composition.
        """
        return self.compose(other)
    
    def __getitem__(self, args):
        r"""
        Slicing operator, for convenience.
        """
        return type(self)(self.linear[args])
    
    def __repr__(self):
        return f"{type(self).__name__}(linear={self.linear.__repr__()})"
    
    def clone(self):
        r"""
        Returns:
            A copy of the transformation (useful to avoid aliasing issues).
        """
        return type(self)(self.linear.clone())

class Orthonormal(Linear):
    r"""
    An orthogonal transformation represented by an orthonormal matrix :math:`M \in \mathcal{M}_{D,D}(\mathbb{R})`,
    transforming a point :math:`x \in \mathbb{R}^D` into :math:`M x`.

    :var linear: (...xDxD tensor): batch of matrices :math:`M` specifying the transformations considered.
    """
    def __init__(self, linear):
        assert linear.shape[-1] == linear.shape[-2], "Expecting same dimensions for input and output."
        super().__init__(linear)

    def linear_inverse(self):
        return self.linear.transpose(-1,-2)

    def linear_normalize(self):
        r"""
        Returns:
            Linear transformation normalized to an orthonormal matrix (...xDxD tensor).
        """        
        return roma.mappings.procrustes(self.linear)
    
class Rotation(Orthonormal):
    r"""
    A rotation represented by a rotation matrix :math:`R \in \mathcal{M}_{D,D}(\mathbb{R})`,
    transforming a point :math:`x \in \mathbb{R}^D` into :math:`R x`.

    :var linear: (...xDxD tensor): batch of matrices :math:`R` defining the rotation.
    """
    def __init__(self, linear):
        super().__init__(linear)

    def linear_normalize(self):
        r"""
        Returns:
            Linear transformation normalized to a rotation matrix (...xDxD tensor).
        """
        return roma.mappings.special_procrustes(self.linear)
    
class RotationUnitQuat(Linear):
    r"""
    A 3D rotation represented by a unit quaternion.
    
    :var linear: (...x4 tensor, XYZW convention): batch of unit quaternions defining the rotation.

    Note:
        Quaternions are assumed to be of unit norm, for all internal operations.
        Use :code:`normalize()` if needed.
    """
    def __init__(self, linear):
        self.linear = linear

    def linear_compose(self, other):
        return roma.utils.quat_product(self.linear, other.linear)
    
    def linear_apply(self, v):
        return roma.utils.quat_action(self.linear, v, is_normalized=True)

    def linear_inverse(self):
        return roma.utils.quat_conjugation(self.linear)
                                       
    def linear_normalize(self):
        r"""
        Returns:
            Normalized unit quaternion (...x4 tensor).
        """
        unitquat = self.linear / torch.norm(self.linear, dim=-1, keepdim=True)
        return unitquat

class _BaseAffine:
    r"""
    Abstract base class representing an affinity transformation.
    """
    def __init__(self, linear, translation):
        self.linear = linear
        self.translation = translation

    def compose(self, other):
        linear = self.linear_compose(other)
        translation = self.linear_apply(other.translation) + self.translation
        return type(self)(linear, translation)
    
    def inverse(self):
        inv_linear = self.linear_inverse()
        res = type(self)(inv_linear, self.translation)
        res.translation = -res.linear_apply(self.translation)
        return res
    
    def apply(self, v):
        return self.linear_apply(v) + self.translation
    
    def normalize(self):
        linear = self.linear_normalize()
        return type(self)(linear, self.translation)
    
    def __getitem__(self, args):
        r"""
        Slicing operator, for convenience.
        """
        return type(self)(self.linear[args], self.translation[args])
    
    def squeeze(self, dim):
        r"""
        Return a view of the transformation in which a batch dimension equal to 1 has been squeezed.

        :var dim: positive integer: The dimension to squeeze.
        """
        assert dim >= 0, "Only positive dimensions are supported to avoid ambiguities."
        assert self.linear.shape[dim] == self.translation.shape[dim] == 1, ""
        return type(self)(self.linear.squeeze(dim), self.translation.squeeze(dim))
    
    def __len__(self):
        return len(self.linear)
    
    def __repr__(self):
        return f"{type(self).__name__}(linear={self.linear.__repr__()}, translation={self.translation.__repr__()})"

    def as_tuple(self):
        r"""
        Returns:
            a tuple of tensors containing the linear and translation parts of the transformation respectively.
        """
        return self.linear, self.translation
    
    def clone(self):
        r"""
        Returns:
            A copy of the transformation (useful to avoid aliasing issues).
        """
        return type(self)(self.linear.clone(), self.translation.clone())
    

    def to_homogeneous(self, output=None):
        r"""
        Args:
            output (...x(D+1)x(C+1) tensor or None): optional tensor in which to store the result.

        Returns:
            A ...x(D+1)x(C+1) tensor of homogeneous matrices representing the transformation, normalized with a last row equal to (0,...,0,1).
        """
        batch_shape, D, C = self.linear.shape[:-2], self.linear.shape[-2], self.linear.shape[-1]
        output_shape = batch_shape + (D+1,C+1)
        if output is None:
            output = torch.empty(output_shape, device=self.translation.device, dtype=self.translation.dtype)
        else:
            assert output.shape == output_shape
        output[...,:D,:C] = self.linear
        output[...,:D,C] = self.translation
        output[...,D,:C] = 0.0
        output[...,D,C] = 1.0
        return output

    @classmethod
    def from_homogeneous(class_object, matrix):
        r"""
        Instantiate a new transformation from an input homogeneous (D+1)x(C+1) matrix.
        The input matrix is assumed to be normalized and to satisfy the properties of the transformation. No checks are performed.

        Args:
            matrix (...x(D+1)x(C+1) tensor): tensor of transformations expressed in homogeneous coordinates, normalized with a last row equal to (0,...,0,1).

        Returns:
            The corresponding transformation.
        """
        H1, H2 = matrix.shape[-2:]
        D = H1 - 1
        C = H2 - 1
        linear = matrix[...,:D, :C]
        translation = matrix[...,:D, C]
        return class_object(linear, translation)
                                       
class Affine(_BaseAffine, Linear):
    r"""
    An affine transformation represented by a linear and a translation part.

    :var linear: (...xCxD tensor): batch of matrices specifying the linear part.
    :var translation: (...xD tensor or None): batch of matrices specifying the translation part.
    """
    def __init__(self, linear, translation):
        if translation is None:
            # Set a default null translation.
            translation = torch.zeros(linear.shape[:-2] + (linear.shape[-1],), dtype=linear.dtype, device=linear.device)
        assert translation.shape[-1] == linear.shape[-2], "Incompatible linear and translation dimensions."
        assert len(linear.shape[:-2]) == len(translation.shape[:-1]), "Batch dimensions should be broadcastable."
        _BaseAffine.__init__(self, linear, translation)


class Isometry(Affine, Orthonormal):
    r"""
    An isometric transformation represented by an orthonormal and a translation part.

    :var linear: (...xDxD tensor or None): batch of matrices specifying the linear part.
    :var translation: (...xD tensor or None): batch of matrices specifying the translation part.
    """
    def __init__(self, linear, translation):
        if linear is None:
            # Set a default identity linear part.
            batch_dims = translation.shape[:-1]
            D = translation.shape[-1]
            if len(batch_dims) == 0:
                linear = torch.eye(D, dtype=translation.dtype, device=translation.device)
            else:
                linear = torch.eye(D, dtype=translation.dtype, device=translation.device)[[None] * len(batch_dims)].expand(batch_dims + (-1,-1))
        else:
            assert linear.shape[-1] == linear.shape[-2], "Expecting same dimensions for input and output."
        Affine.__init__(self, linear, translation)

    @classmethod
    def Identity(cls, *args, **kwargs):
        warnings.warn(
            "Isometry.Identity is deprecated and will be removed in a future release."
            "Use Isometry.identity instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return cls.identity(*args, **kwargs)

    @classmethod
    def identity(cls, dim, batch_shape=tuple(), dtype=torch.float32, device=None):
        r"""
        Return a default identity transformation.

        :var dim: (strictly positive integer): dimension of the space in which the transformation operates (e.g. `dim=3` for 3D transformations).
        :var batch_shape: (tuple): batch dimensions considered.
        """
        translation = torch.zeros(batch_shape + (dim,), dtype=dtype, device=device)
        return cls(linear=None, translation=translation)
    
    @classmethod
    def identity_like(cls, other):
        r"""
        Return an identity transformation with the same batch shape, type and device as the input transformation.

        :var other: (Isometry): input transformation.
        """
        return cls.identity(dim=other.translation.shape[-1], batch_shape=other.translation.shape[:-1], dtype=other.linear.dtype, device=other.linear.device)

class Rigid(Isometry, Rotation):
    r"""
    A rigid transformation represented by an rotation and a translation part.

    :var linear: (...xDxD tensor or None): batch of matrices specifying the linear part.
    :var translation: (...xD tensor or None): batch of matrices specifying the translation part.
    """
    def __init__(self, linear, translation):
        Isometry.__init__(self, linear, translation)

    def to_rigidunitquat(self):
        r"""
        Returns the corresponding RigidUnitQuat transformation.

        Note:
            Original and resulting transformations share the same translation tensor. Be careful in case of in-place modifications.
        """
        return RigidUnitQuat(roma.rotmat_to_unitquat(self.linear), self.translation)

class RigidUnitQuat(_BaseAffine, RotationUnitQuat):
    r"""
    A rigid transformation represented by a unit quaternion and a translation part.

    :var linear: (...x4 tensor): batch of unit quaternions defining the rotation.
    :var translation: (...x3 tensor): batch of matrices specifying the translation part.    

    Note:
        Quaternions are assumed to be of unit norm, for all internal operations.
        Use the :code:`normalize()` method if needed.
    """
    def __init__(self, linear, translation):
        assert linear.shape[-1] == 4 and translation.shape[-1] == 3, "Expecting respectively a ...x4 quaternion vector and a ...x3 translation vector"
        assert len(linear.shape[:-1]) == len(translation.shape[:-1]), "Batch dimensions should be broadcastable."
        _BaseAffine.__init__(self, linear, translation)

    def to_homogeneous(self, output=None):
        r"""
        Args:
            output (...x4x4 tensor or None): tensor in which to store the result (optional).

        Returns:
            A ...x4x4 tensor of homogeneous matrices representing the transformation, normalized with a last row equal to (0,...,0,1).
        """
        batch_shape = self.translation.shape[:-1]
        output_shape = batch_shape + (4,4)
        if output is None:
            output = torch.zeros(output_shape, device=self.translation.device, dtype=self.translation.dtype)
        else:
            assert output_shape == output_shape
        output[...,:3,:3] = roma.unitquat_to_rotmat(self.linear)
        output[...,:3,3] = self.translation
        # Set the homogeneous line
        # Note: this is redundant with zeros initialization, but .
        output[...,3,:3] = 0.0
        output[...,3,3] = 1.0
        return output

    @staticmethod
    def from_homogeneous(matrix):
        r"""
        Instantiate a new transformation from an input homogeneous (D+1)x(D+1) matrix.

        Args:
            matrix (...x(D+1)x(D+1) tensor): tensor of transformations expressed in homogeneous coordinates, normalized with a last row equal to (0,...,0,1).

        Returns:
            The corresponding transformation.

        Note:
            - The input matrix is not tested to ensure that it satisfies the required properties of the transformation.
            - Components of the resulting transformation may consist in views of the input matrix. Be careful if you intend to modify it in-place.
            
        """
        H1, H2 = matrix.shape[-2:]
        assert H1 == H2
        D = H1 - 1
        linear = roma.rotmat_to_unitquat(matrix[...,:D, :D])
        translation = matrix[...,:D, D]
        return RigidUnitQuat(linear, translation)
    
    def to_rigid(self):
        r"""
        Returns the corresponding Rigid transformation.

        Note:
            Original and resulting transformations share the same translation tensor. Be careful in case of in-place modifications.
        """
        return Rigid(roma.unitquat_to_rotmat(self.linear), self.translation)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="roma",
    version="1.5.4",
    author="Romain Brégier",
    author_email="romain.bregier@naverlabs.com",
    description="A lightweight library to deal with 3D rotations in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/naver/roma",
    packages=["roma"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


# RoMa
# Copyright (c) 2020 NAVER Corp.
# 3-Clause BSD License.
import unittest
import torch
import roma
import numpy as np
from test.utils import is_close
import itertools

device = torch.device(0) if torch.cuda.is_available() else torch.device('cpu')

class TestEuler(unittest.TestCase):
    def test_euler(self):
        batch_shape = torch.Size((3,2))
        x = torch.randn(batch_shape)
        y = torch.randn(batch_shape)
        q = roma.euler_to_unitquat('xy', (x, y))
        self.assertTrue(q.shape == batch_shape + (4,))

    def test_euler_unitquat_consistency(self):
        device = torch.device(0) if torch.cuda.is_available() else torch.device('cpu')
        dtype = torch.float64
        for degrees in (True, False):
            for batch_shape in [tuple(),
                                torch.Size((30,)),
                                torch.Size((50,60))]:
                for intrinsics in (True, False):
                    for convention in ["".join(permutation) for permutation in itertools.permutations('xyz')] + ["xyx", "xzx", "yxy", "yzy", "zxz", "zyz"]:
                        if intrinsics:
                            convention = convention.upper()
                        q = roma.random_unitquat(batch_shape, device=device, dtype=dtype)
                        angles = roma.unitquat_to_euler(convention, q, degrees=degrees, as_tuple=True)
                        self.assertTrue(len(angles) == 3)
                        self.assertTrue(all([angle.shape == batch_shape for angle in angles]))
                        if degrees:
                            self.assertTrue(all([torch.all(angle > -180.) and torch.all(angle <= 180) for angle in angles]))
                        else:
                            self.assertTrue(all([torch.all(angle > -np.pi) and torch.all(angle <= np.pi) for angle in angles]))
                        q1 = roma.euler_to_unitquat(convention, angles, degrees=degrees)
                        self.assertTrue(torch.all(roma.unitquat_geodesic_distance(q, q1) < 1e-6))

    def test_euler_rotvec_consistency(self):
        device = torch.device(0) if torch.cuda.is_available() else torch.device('cpu')
        dtype = torch.float64
        for degrees in (True, False):
            for batch_shape in [tuple(),
                                torch.Size((30,)),
                                torch.Size((50,60))]:
                for intrinsics in (True, False):
                    for convention in ["".join(permutation) for permutation in itertools.permutations('xyz')] + ["xyx", "xzx", "yxy", "yzy", "zxz", "zyz"]:
                        if intrinsics:
                            convention = convention.upper()
                        q = roma.random_rotvec(batch_shape, device=device, dtype=dtype)
                        angles = roma.rotvec_to_euler(convention, q, degrees=degrees, as_tuple=True)
                        self.assertTrue(len(angles) == 3)
                        self.assertTrue(all([angle.shape == batch_shape for angle in angles]))
                        if degrees:
                            self.assertTrue(all([torch.all(angle > -180.) and torch.all(angle <= 180) for angle in angles]))
                        else:
                            self.assertTrue(all([torch.all(angle > -np.pi) and torch.all(angle <= np.pi) for angle in angles]))
                        q1 = roma.euler_to_rotvec(convention, angles, degrees=degrees)
                        self.assertTrue(torch.all(roma.rotvec_geodesic_distance(q, q1) < 1e-6))

    def test_euler_rotmat_consistency(self):
        device = torch.device(0) if torch.cuda.is_available() else torch.device('cpu')
        dtype = torch.float64
        for degrees in (True, False):
            for batch_shape in [tuple(),
                                torch.Size((30,)),
                                torch.Size((50,60))]:
                for intrinsics in (True, False):
                    for convention in ["".join(permutation) for permutation in itertools.permutations('xyz')] + ["xyx", "xzx", "yxy", "yzy", "zxz", "zyz"]:
                        if intrinsics:
                            convention = convention.upper()
                        q = roma.random_rotmat(batch_shape, device=device, dtype=dtype)
                        angles = roma.rotmat_to_euler(convention, q, degrees=degrees, as_tuple=True)
                        self.assertTrue(len(angles) == 3)
                        self.assertTrue(all([angle.shape == batch_shape for angle in angles]))
                        if degrees:
                            self.assertTrue(all([torch.all(angle > -180.) and torch.all(angle <= 180) for angle in angles]))
                        else:
                            self.assertTrue(all([torch.all(angle > -np.pi) and torch.all(angle <= np.pi) for angle in angles]))
                        q1 = roma.euler_to_rotmat(convention, angles, degrees=degrees)
                        self.assertTrue(torch.all(roma.rotmat_geodesic_distance(q, q1) < 1e-6))

    def test_euler_backward(self):
        for intrinsics in (True, False):
            for convention in ["".join(permutation) for permutation in itertools.permutations('xyz')] + ["xyx", "xzx", "yxy", "yzy", "zxz", "zyz"]:
                if intrinsics:
                    convention = convention.upper()
                    rotvec = torch.randn(3, requires_grad=True)
                    angles = roma.rotvec_to_euler('xyz', rotvec)
                    sum(angles).backward()

    def test_euler_tensor(self):
        r"""
        Test that Euler conversion methods support both list and tensor inputs.
        """
        batch_shape = (10,34)
        dtype = torch.float64
        q = roma.random_unitquat(batch_shape, device=device, dtype=dtype)
        convention = 'xyz'
        angles = roma.unitquat_to_euler(convention, q, as_tuple=True)
        angles_tensor = roma.unitquat_to_euler(convention, q)
        assert type(angles) == tuple
        assert type(angles_tensor) == torch.Tensor
        q1 = roma.euler_to_unitquat(convention, angles)
        q2 = roma.euler_to_unitquat(convention, angles_tensor)
        self.assertTrue(torch.all(roma.rotmat_geodesic_distance(q1, q2) < 1e-6))


if __name__ == "__main__":
    unittest.main()